/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;

import javax.swing.JFileChooser;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.KeyStroke;

import vispaca.data.DataManager;
import vispaca.data.DataParser;
import vispaca.data.DataTable;


/**
 * VisPaca main panel.
 */
public class MainPanel extends JPanel {

	private static final long serialVersionUID = -8973474815065008130L;

	public static final String VIS_PARALLEL = "Parallel";
	public static final String VIS_SCATTER = "Scatter";
	public static final String VIS_TABLE = "Table";

	private VDataTable tablePanel = new VDataTable();
	private VFilterPanel filterPanel = new VFilterPanel();

	private DataManager dataManager;

	private Frame frame;


	public MainPanel(Frame frame) {
		this.frame = frame;

		setLayout(new BorderLayout());

		JMenuBar menuBar = createMenuBar();

		JTabbedPane tabPane = new JTabbedPane();
		tabPane.addTab("Table", new JScrollPane(tablePanel));
		tabPane.addTab("Filter", new JScrollPane(filterPanel));

		add(menuBar, BorderLayout.NORTH);
		add(tabPane, BorderLayout.CENTER);
	}


	/**
	 * @return The menu bar for this panel.
	 */
	private JMenuBar createMenuBar() {
		JMenuBar menuBar = new JMenuBar();

		JMenu fileMenu = new JMenu("File");

		JMenuItem openMenuItem = new JMenuItem("Open...");
		openMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, KeyEvent.CTRL_MASK));
		openMenuItem.addActionListener(new OpenListener());
		fileMenu.add(openMenuItem);

		JMenuItem exitMenuItem = new JMenuItem("Exit");
		exitMenuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, KeyEvent.CTRL_MASK));
		exitMenuItem.addActionListener(new ExitListener());
		fileMenu.add(exitMenuItem);

		JMenu winMenu = new JMenu("Window");
		NewWindowListener winListener = new NewWindowListener();

		JMenuItem winParallelMenuItem = new JMenuItem(VIS_PARALLEL);
		winParallelMenuItem.addActionListener(winListener);
		winMenu.add(winParallelMenuItem);

		JMenuItem winScatterMenuItem = new JMenuItem(VIS_SCATTER);
		winScatterMenuItem.addActionListener(winListener);
		winMenu.add(winScatterMenuItem);

		JMenuItem winTableMenuItem = new JMenuItem(VIS_TABLE);
		winTableMenuItem.addActionListener(winListener);
		winMenu.add(winTableMenuItem);

		menuBar.add(fileMenu);
		menuBar.add(winMenu);

		return menuBar;
	}


	/**
	 * Listener for opening files.
	 */
	private class OpenListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent actionEvent) {
			JFileChooser fileChooser = new JFileChooser();
			int status = fileChooser.showOpenDialog(MainPanel.this);

			if (status != JFileChooser.APPROVE_OPTION) {
				return;
			}

			File file = fileChooser.getSelectedFile();

			if (!file.canRead()) {
				JOptionPane.showMessageDialog(
						MainPanel.this,
						file + " is unreadable.",
						"Unable to Read File",
						JOptionPane.ERROR_MESSAGE
				);
			}

			try {
				InputStream inputStream = new FileInputStream(file);
				DataTable dataTable = DataParser.parse(inputStream);

				if (dataManager != null) {
					dataManager.destroyPerformed();
				}

				dataManager = new DataManager(dataTable);

				tablePanel.setDataManager(dataManager);
				filterPanel.setDataManager(dataManager);

			} catch (Exception e) {
				e.printStackTrace();

				JOptionPane.showMessageDialog(
						MainPanel.this,
						e,
						"Unable to Parse File",
						JOptionPane.ERROR_MESSAGE
				);
			}
		}
	}


	/**
	 * Listener for exiting the program.
	 */
	private class ExitListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			System.exit(0);
		}
	}


	/**
	 * Creates new windows.
	 */
	private class NewWindowListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			if (dataManager == null) {
				return;
			}

			String actionCommand = e.getActionCommand();
			VFrame vFrame = new VFrame(dataManager);

			switch (actionCommand) {
			case VIS_PARALLEL:
				vFrame.setVPanel(new VParallelPanel(vFrame, dataManager));
				vFrame.setTitle("VisPaca - " + VIS_PARALLEL);
				break;
			case VIS_SCATTER:
				vFrame.setVPanel(new VScatterPanel(vFrame, dataManager));
				vFrame.setTitle("VisPaca - " + VIS_SCATTER);
				break;
			case VIS_TABLE:
				vFrame.setVPanel(new VTablePanel(vFrame, dataManager));
				vFrame.setTitle("VisPaca - " + VIS_TABLE);
				break;
			}

			vFrame.setLocationRelativeTo(frame);
			vFrame.setVisible(true);
		}
	}
}
