/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Objects;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import vispaca.data.DataTable;
import vispaca.misc.ParallelConfig;
import vispaca.misc.PanelUtil;


/**
 * Allows for the selection of visible columns and their order.
 */
public class VAttributeSelectionPanel extends JPanel {

	private static final long serialVersionUID = -6206894424484537813L;

	private JList<ModelEntry> list = new JList<>();
	private DefaultListModel<ModelEntry> listModel = new DefaultListModel<>();

	private JComboBox<ModelEntry> comboBox = new JComboBox<>();
	private DefaultComboBoxModel<ModelEntry> comboBoxModel = new DefaultComboBoxModel<>();


	public VAttributeSelectionPanel(DataTable dataTable, ParallelConfig config) {
		Objects.requireNonNull(dataTable);
		Objects.requireNonNull(config);

		setLayout(new BorderLayout());
		setBorder(BorderFactory.createTitledBorder("Attributes"));

		list.setModel(listModel);
		comboBox.setModel(comboBoxModel);

		populateComboBox(dataTable);
		populateList(dataTable, config);

		list.setVisibleRowCount(7);

		JPanel attributePanel = new JPanel(new BorderLayout());
		attributePanel.add(new JScrollPane(list), BorderLayout.CENTER);

		JPanel controlPanel = getControlPanel();

		add(attributePanel, BorderLayout.CENTER);
		add(controlPanel, BorderLayout.SOUTH);
	}


	/**
	 * @return All selected attributes.
	 */
	public int[] getAttributes() {
		int[] attributes = new int[listModel.getSize()];

		for (int i = 0; i < attributes.length; i++) {
			attributes[i] = listModel.get(i).attribute;
		}

		return attributes;
	}


	/**
	 * @return Panel for adding, removing and reordering attributes.
	 */
	private JPanel getControlPanel() {
		JPanel controlPanel = new JPanel(new GridBagLayout());
		GridBagConstraints constraints = PanelUtil.getHorizontalConstraints();

		JButton addButton = new JButton("+");
		addButton.addActionListener(new AddButtonListener());

		JButton removeButton = new JButton("-");
		removeButton.addActionListener(new RemoveButtonListener());

		JButton upButton = new JButton("\u2191");
		upButton.addActionListener(new UpButtonListener());

		JButton downButton = new JButton("\u2193");
		downButton.addActionListener(new DownButtonListener());

		controlPanel.add(comboBox, constraints);
		controlPanel.add(addButton, constraints);
		controlPanel.add(removeButton, constraints);
		controlPanel.add(upButton, constraints);
		controlPanel.add(downButton, constraints);

		return controlPanel;
	}


	/**
	 * Adds all numerical attributes to the combo box.
	 */
	private void populateComboBox(DataTable dataTable) {
		int attributeCount = dataTable.getAttributeCount();

		for (int i = 0; i < attributeCount; i++) {
			if (!Number.class.isAssignableFrom(dataTable.getAttributeType(i))) {
				continue;
			}

			ModelEntry entry = new ModelEntry(dataTable.getAttributeName(i), i);
			comboBoxModel.addElement(entry);
		}

		comboBox.setSelectedIndex(0);
	}


	/**
	 * Adds all currently visible attributes to the list.
	 */
	private void populateList(DataTable dataTable, ParallelConfig config) {
		int axisCount = config.getAxisCount();

		for (int i = 0; i < axisCount; i++) {
			int attribute = config.getAxis(i).getAttribute();

			if (attribute < 0) {
				continue;
			}

			ModelEntry entry = new ModelEntry(dataTable.getAttributeName(attribute), attribute);
			listModel.addElement(entry);
		}
	}


	/**
	 * Simple entry class for the combo box and list.
	 */
	private class ModelEntry {
		String name;
		int attribute;

		public ModelEntry(String name, int attribute) {
			this.name = name;
			this.attribute = attribute;
		}

		@Override
		public String toString() {
			return name;
		}
	}


	/**
	 * Adds a new attribute to the list when the add button is clicked.
	 */
	private class AddButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			listModel.addElement((ModelEntry) comboBoxModel.getSelectedItem());
		}
	}


	/**
	 * Removes an attribute when the remove button is clicked.
	 */
	private class RemoveButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int selectedRow = list.getSelectedIndex();

			if (selectedRow < 0) {
				return;
			}

			listModel.remove(selectedRow);
		}
	}


	/**
	 * Reorders the selected attributes when the up button is clicked.
	 */
	private class UpButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int selectedRow = list.getSelectedIndex();

			if (selectedRow <= 0) {
				return;
			}

			ModelEntry topEntry = listModel.get(selectedRow - 1);
			ModelEntry bottomEntry = listModel.get(selectedRow);

			listModel.setElementAt(topEntry, selectedRow);
			listModel.setElementAt(bottomEntry, selectedRow - 1);

			list.setSelectedIndex(selectedRow - 1);
		}
	}


	/**
	 * Reorders the selected attributes when the down button is clicked.
	 */
	private class DownButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int selectedRow = list.getSelectedIndex();

			if (listModel.getSize() <= selectedRow + 1) {
				return;
			}

			ModelEntry topEntry = listModel.get(selectedRow);
			ModelEntry bottomEntry = listModel.get(selectedRow + 1);

			listModel.setElementAt(topEntry, selectedRow + 1);
			listModel.setElementAt(bottomEntry, selectedRow);

			list.setSelectedIndex(selectedRow + 1);
		}
	}
}
