/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.util.Comparator;
import java.util.Objects;

import javax.swing.JTable;
import javax.swing.RowFilter;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import vispaca.data.DataElement;
import vispaca.data.DataManager;
import vispaca.data.DataTable;
import vispaca.event.DestroyListener;
import vispaca.event.UpdateListener;
import vispaca.misc.ColorDefiner;


/**
 * Displays data as a table.
 */
public class VDataTable extends JTable implements DestroyListener {

	private static final long serialVersionUID = 2109086675573674127L;

	private static final TableCellRenderer NEUTRAL = new MyTableCellRenderer(ColorDefiner.NEUTRAL);
	private static final TableCellRenderer TOUCHED = new MyTableCellRenderer(ColorDefiner.TOUCHED);
	private static final TableCellRenderer HIGHLIGHTED = new MyTableCellRenderer(ColorDefiner.HIGHLIGHTED);

	private DataManager dataManager = null;
	private DataTable dataTable = null;

	private UpdateListener updateListener = new MyUpdateListener();

	private MyTableModel tableModel;


	public VDataTable() {
		setAutoCreateRowSorter(true);
		addMouseMotionListener(new MyMouseMotionListener());
		addMouseListener(new MyMouseListener());
	}


	/**
	 * @param dataManager The data manager to use.
	 */
	public void setDataManager(DataManager dataManager) {
		Objects.requireNonNull(dataManager);

		if (this.dataManager != null) {
			this.dataManager.removeUpdateListener(updateListener);
		}

		this.dataManager = dataManager;
		this.dataTable = dataManager.getDataTable();

		tableModel = new MyTableModel(dataManager);

		TableRowSorter<MyTableModel> tableRowSorter = new TableRowSorter<>(tableModel);
		tableRowSorter.setRowFilter(new MyRowFilter());

		int attributeCount = dataTable.getAttributeCount();

		/* Proper sorting depending on attribute type. */
		for (int i = 0; i < attributeCount; i++) {
			Class<?> type = dataTable.getAttributeType(i);

			if (String.class.isAssignableFrom(type)) {
				tableRowSorter.setComparator(i, new Comparator<String>() {
					@Override
					public int compare(String o1, String o2) {
						return o1.compareTo(o2);
					}
				});

			} else if (Number.class.isAssignableFrom(type)) {
				tableRowSorter.setComparator(i, new Comparator<String>() {
					@Override
					public int compare(String o1, String o2) {
						double d1 = Double.parseDouble(o1);
						double d2 = Double.parseDouble(o2);

						if (d1 == d2) {
							return 0;
						} else if (d1 < d2) {
							return -1;
						} else {
							return 1;
						}
					}
				});
			}
		}

		setModel(tableModel);
		setRowSorter(tableRowSorter);

		dataManager.addUpdateListener(updateListener);

		repaint();
	}


	@Override
	public TableCellRenderer getCellRenderer(int row, int column) {
		DataElement element = dataTable.getElement(getRowSorter().convertRowIndexToModel(row));

		if (dataManager.isHighlighted(element)) {
			return HIGHLIGHTED;
		} else if (dataManager.isTouched(element)) {
			return TOUCHED;
		} else {
			return NEUTRAL;
		}
	}


	@Override
	public void destroyPerformed() {
		dataManager.removeUpdateListener(updateListener);
		dataManager = null;
		dataTable = null;
	}


	/**
	 * Maps the DataTable to JTable.
	 */
	private class MyTableModel extends AbstractTableModel {

		private static final long serialVersionUID = -2895189514903015693L;

		private DataTable dataTable;


		public MyTableModel(DataManager dataManager) {
			this.dataTable = dataManager.getDataTable();
		}

		@Override
		public Object getValueAt(int rowIndex, int columnIndex) {
			return dataTable.getElement(rowIndex).getString(columnIndex);
		}

		@Override
		public int getRowCount() {
			return dataTable.getElementCount();
		}

		@Override
		public int getColumnCount() {
			return dataTable.getAttributeCount();
		}

		@Override
		public String getColumnName(int column) {
			return dataTable.getAttributeName(column);
		}

		@Override
		public boolean isCellEditable(int rowIndex, int columnIndex) {
			return false;
		}
	}


	/**
	 * Overrides text colors.
	 */
	private static class MyTableCellRenderer extends DefaultTableCellRenderer {

		private static final long serialVersionUID = 2011542865712485108L;

		private Color color;


		public MyTableCellRenderer(Color color) {
			this.color = color;
		}

		@Override
		public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
			Component component = super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);

			component.setForeground(color);
			component.setBackground(ColorDefiner.BACKGROUND);

			return component;
		}
	}


	/**
	 * Handles filtering of elements.
	 */
	public class MyRowFilter extends RowFilter<TableModel, Integer> {
		@Override
		public boolean include(javax.swing.RowFilter.Entry<? extends TableModel, ? extends Integer> entry) {
			if (dataManager != null && dataTable != null) {
				int row = entry.getIdentifier();
				DataElement element = dataTable.getElement(row);
				return dataManager.isVisible(element);
			}

			return true;
		}
	}


	/**
	 * Touches elements under the cursor.
	 */
	private class MyMouseMotionListener extends MouseMotionAdapter {
		@Override
		public void mouseMoved(MouseEvent e) {
			if (dataManager != null && dataTable != null) {
				int row = rowAtPoint(e.getPoint());

				if (row < 0) {
					dataManager.clearTouched();
					return;
				}

				row = getRowSorter().convertRowIndexToModel(row);
				DataElement element = dataTable.getElement(row);

				if (!dataManager.isTouched(element)) {
					dataManager.setTouched(element);
					repaint();
				}
			}
		}
	}


	/**
	 * Highlights clicked elements.
	 */
	private class MyMouseListener extends MouseAdapter {
		@Override
		public void mouseClicked(MouseEvent e) {
			if (dataManager != null && dataTable != null) {
				int row = rowAtPoint(e.getPoint());

				if (row < 0) {
					return;
				}

				row = getRowSorter().convertRowIndexToModel(row);
				DataElement element = dataTable.getElement(row);

				if (e.isControlDown()) {
					if (dataManager.isHighlighted(element)) {
						dataManager.removeHighlight(element);
					} else {
						dataManager.addHighlight(element);
					}
				} else {
					dataManager.clearHighlight();
					dataManager.addHighlight(element);
				}

				repaint();
			}
		}
	}


	/**
	 * Repaints the table on update.
	 */
	private class MyUpdateListener implements UpdateListener {
		@Override
		public void updatePerformed() {
			if (tableModel != null) {
				tableModel.fireTableDataChanged();
			}
		}
	}
}
