/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Objects;

import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import vispaca.data.DataManager;
import vispaca.data.DataTable;
import vispaca.misc.AttributeConfig;
import vispaca.misc.PanelUtil;


/**
 * Filters a single column of numbers.
 */
public class VNumberConfigPanel extends JPanel {

	private static final long serialVersionUID = 7505339055530743864L;

	private boolean reverse;

	private AttributeConfig config;

	private JComboBox<ComboBoxEntry> comboBox = new JComboBox<>();
	private DefaultComboBoxModel<ComboBoxEntry> comboBoxModel = new DefaultComboBoxModel<>();

	private JSpinner leftSpinner = new JSpinner();
	private JSpinner rightSpinner = new JSpinner();

	private DataTable dataTable;


	public VNumberConfigPanel(DataManager dataManager, AttributeConfig config, String title) {

		/* Input validation. */
		Objects.requireNonNull(dataManager);
		Objects.requireNonNull(config);
		Objects.requireNonNull(title);

		dataTable = dataManager.getDataTable();

		this.config = config;

		/* Combo box. */
		ComboBoxEntry noneEntry = new ComboBoxEntry("None", Object.class, -1);
		comboBoxModel.addElement(noneEntry);
		comboBoxModel.setSelectedItem(noneEntry);

		int attributeCount = dataTable.getAttributeCount();
		int currentAttribute = config.getAttribute();

		for (int i = 0; i < attributeCount; i++) {
			ComboBoxEntry entry = new ComboBoxEntry(
					dataTable.getAttributeName(i),
					dataTable.getAttributeType(i),
					i
			);

			if (Number.class.isAssignableFrom(entry.type)) {
				comboBoxModel.addElement(entry);
			}

			if (currentAttribute == i) {
				comboBoxModel.setSelectedItem(entry);
			}
		}

		comboBox.setModel(comboBoxModel);

		/* Reset button */
		JButton resetButton = new JButton("Reset");

		/* Events. */
		leftSpinner.addChangeListener(new LeftSpinnerChangeListener());
		rightSpinner.addChangeListener(new RightSpinnerChangeListener());
		comboBox.addActionListener(new ComboBoxActionListener());
		resetButton.addActionListener(new ResetButtonListener());

		/* Panel layout */
		setLayout(new GridBagLayout());
		setBorder(BorderFactory.createTitledBorder(title));
		GridBagConstraints constraints = PanelUtil.getHorizontalConstraints();

		Dimension spinnerSize = new Dimension(
				80,
				leftSpinner.getPreferredSize().height
		);

		leftSpinner.setPreferredSize(spinnerSize);
		rightSpinner.setPreferredSize(spinnerSize);

		add(comboBox, constraints);
		add(leftSpinner, constraints);
		add(new JLabel(" < "), constraints);
		add(rightSpinner, constraints);
		add(resetButton, constraints);

		try {
			initializeSpinners();
		} catch (IllegalArgumentException e) {
			config.init(dataTable, config.getAttribute());
			initializeSpinners();
		}
	}


	/**
	 * Initializes the spinners to a new attribute.
	 */
	private void initializeSpinners() {
		ComboBoxEntry entry = (ComboBoxEntry) comboBoxModel.getSelectedItem();

		if (config.getAttribute() != entry.attribute) {
			config.init(dataTable, entry.attribute);
		}

		double initialMin;
		double initialMax;
		double currentMin;
		double currentMax;

		if (config.getInitialMin() <= config.getInitialMax()) {
			reverse = false;
			initialMin = config.getInitialMin();
			initialMax = config.getInitialMax();
			currentMin = config.getCurrentMin();
			currentMax = config.getCurrentMax();
		} else {
			reverse = true;
			initialMin = config.getInitialMax();
			initialMax = config.getInitialMin();
			currentMin = config.getCurrentMax();
			currentMax = config.getCurrentMin();
		}

		double delta = initialMax - initialMin;

		if (Integer.class.isAssignableFrom(entry.type)) {
			int step = (int) (delta / 20);

			if (step < 1) {
				step = 1;
			}

			leftSpinner.setModel(new SpinnerNumberModel(currentMin, initialMin, initialMax, step));
			((JSpinner.NumberEditor) leftSpinner.getEditor()).getFormat().setMaximumFractionDigits(0);
			rightSpinner.setModel(new SpinnerNumberModel(currentMax, initialMin, initialMax, step));
			((JSpinner.NumberEditor) rightSpinner.getEditor()).getFormat().setMaximumFractionDigits(0);
		} else {
			double step = delta / 20;

			if (step < 1) {
				step = 0.1;
			}

			leftSpinner.setModel(new SpinnerNumberModel(currentMin, initialMin, initialMax, step));
			((JSpinner.NumberEditor) leftSpinner.getEditor()).getFormat().setMinimumFractionDigits(2);
			((JSpinner.NumberEditor) leftSpinner.getEditor()).getFormat().setMaximumFractionDigits(2);
			rightSpinner.setModel(new SpinnerNumberModel(currentMax, initialMin, initialMax, step));
			((JSpinner.NumberEditor) leftSpinner.getEditor()).getFormat().setMinimumFractionDigits(2);
			((JSpinner.NumberEditor) rightSpinner.getEditor()).getFormat().setMaximumFractionDigits(2);
		}
	}


	/**
	 * Updates the config for the left spinner.
	 */
	private class LeftSpinnerChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent e) {
			double value = (Double) leftSpinner.getValue();

			if (reverse) {
				config.setCurrentMax(value);
			} else {
				config.setCurrentMin(value);
			}
		}
	}


	/**
	 * Updates the config for the right spinner.
	 */
	private class RightSpinnerChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent e) {
			double value = (Double) rightSpinner.getValue();

			if (reverse) {
				config.setCurrentMin(value);
			} else {
				config.setCurrentMax(value);
			}
		}
	}


	/**
	 * Initializes spinners on combo box change.
	 */
	private class ComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			initializeSpinners();
		}
	}


	/**
	 * Entity object for the combo box.
	 */
	private class ComboBoxEntry {
		String name;
		Class<?> type;
		int attribute;

		public ComboBoxEntry(String name, Class<?> type, int attribute) {
			this.name = name;
			this.type = type;
			this.attribute = attribute;
		}

		@Override
		public String toString() {
			return name;
		}
	}


	/**
	 * Resets the config and spinners when the reset button is clicked.
	 */
	public class ResetButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			config.init(dataTable, config.getAttribute());
			initializeSpinners();
		}
	}
}
