/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Objects;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import vispaca.data.DataElement;
import vispaca.data.DataManager;
import vispaca.data.DataSelection;
import vispaca.data.DataTable;
import vispaca.event.UpdateMixin;
import vispaca.misc.PanelUtil;


/**
 * Filters a single column of numbers.
 */
public class VNumberFilterPanel extends JPanel {

	private static final long serialVersionUID = 7505339055530743864L;

	private JSpinner minSpinner = new JSpinner();
	private JSpinner maxSpinner = new JSpinner();

	private DataTable dataTable;

	private int attribute;
	private Class<?> type;

	private Filter filter;


	public VNumberFilterPanel(DataManager dataManager, int attribute) {

		/* Input validation. */
		Objects.requireNonNull(dataManager);

		dataTable = dataManager.getDataTable();

		if (attribute < 0 || dataTable.getAttributeCount() <= attribute) {
			throw new IllegalArgumentException("Column " + attribute + " does not exist.");
		}

		type = dataTable.getAttributeType(attribute);
		if (!Number.class.isAssignableFrom(type)) {
			throw new IllegalArgumentException("Attribute type must be subclass of Number.");
		}

		this.attribute = attribute;

		/* Widget and filter initialization. */
		initializeSpinners();
		JButton resetButton = new JButton("Reset");
		resetButton.addActionListener(new ResetButtonListener());

		filter = new Filter();
		dataManager.addFilter(filter);
		minSpinner.addChangeListener(filter);
		maxSpinner.addChangeListener(filter);

		/* Panel layout */
		setLayout(new GridBagLayout());
		setBorder(BorderFactory.createTitledBorder(dataTable.getAttributeName(attribute)));
		GridBagConstraints constraints = PanelUtil.getHorizontalConstraints();

		Dimension spinnerSize = new Dimension(
				80,
				minSpinner.getPreferredSize().height
		);

		minSpinner.setPreferredSize(spinnerSize);
		maxSpinner.setPreferredSize(spinnerSize);

		add(minSpinner, constraints);
		add(new JLabel(" < "), constraints);
		add(maxSpinner, constraints);
		add(resetButton, constraints);
	}


	/**
	 * Sets the spinner values and limits.
	 */
	private void initializeSpinners() {
		double min = Double.MAX_VALUE;
		double max = Double.MIN_VALUE;
		int elementCount = dataTable.getElementCount();

		for (int i = 0; i < elementCount; i++) {
			double value = dataTable.getElement(i).getDouble(attribute);

			if (value < min) {
				min = value;
			}

			if (max < value) {
				max = value;
			}
		}

		double delta = max - min;

		if (Integer.class.isAssignableFrom(type)) {
			int step = (int) (delta / 20);

			if (step < 1) {
				step = 1;
			}

			minSpinner.setModel(new SpinnerNumberModel(min, min, max, step));
			((JSpinner.NumberEditor) minSpinner.getEditor()).getFormat().setMaximumFractionDigits(0);
			maxSpinner.setModel(new SpinnerNumberModel(max, min, max, step));
			((JSpinner.NumberEditor) maxSpinner.getEditor()).getFormat().setMaximumFractionDigits(0);

		} else {
			double step = delta / 20;

			if (step < 1) {
				step = 0.1;
			}

			minSpinner.setModel(new SpinnerNumberModel(min, min, max, step));
			((JSpinner.NumberEditor) minSpinner.getEditor()).getFormat().setMinimumFractionDigits(2);
			((JSpinner.NumberEditor) minSpinner.getEditor()).getFormat().setMaximumFractionDigits(2);
			maxSpinner.setModel(new SpinnerNumberModel(max, min, max, step));
			((JSpinner.NumberEditor) minSpinner.getEditor()).getFormat().setMinimumFractionDigits(2);
			((JSpinner.NumberEditor) maxSpinner.getEditor()).getFormat().setMaximumFractionDigits(2);
		}
	}


	/**
	 * Applies the filter.
	 */
	private class Filter extends UpdateMixin implements ChangeListener, DataSelection {
		@Override
		public void stateChanged(ChangeEvent e) {
			fireUpdateListeners();
		}

		@Override
		public boolean isSelected(DataElement element) {
			double value = element.getDouble(attribute);
			double min = (double) minSpinner.getValue();
			double max = (double) maxSpinner.getValue();

			return value < min || max < value;
		}
	}


	/**
	 * Resets the spinners when the reset button is clicked.
	 */
	private class ResetButtonListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			initializeSpinners();
			filter.fireUpdateListeners();
		}
	}
}
