/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import javax.swing.JDialog;
import javax.swing.ToolTipManager;

import vispaca.data.DataElement;
import vispaca.data.DataManager;
import vispaca.data.DataTable;
import vispaca.event.UpdateListener;
import vispaca.misc.AttributeConfig;
import vispaca.misc.ColorDefiner;
import vispaca.misc.CoordinateMapper;
import vispaca.misc.DataElementUtil;
import vispaca.misc.Line;
import vispaca.misc.ParallelConfig;
import vispaca.misc.ParallelRulerUtil;


/**
 * Panel displaying a histogram.
 */
public class VParallelPanel extends VPanel implements UpdateListener {

	private static final long serialVersionUID = 7753018528178808238L;

	private Map<Line, Set<DataElement>> lines = new HashMap<>();

	private ParallelConfig config = new ParallelConfig();
	private AttributeConfig colorConfig;

	private CoordinateMapper[] axisMappers = new CoordinateMapper[0];
	private CoordinateMapper xMapper = new CoordinateMapper();
	private CoordinateMapper colorMapper = new CoordinateMapper();

	private ParallelRulerUtil rulerUtil = new ParallelRulerUtil();

	private DataManager dataManager;
	private DataTable dataTable;


	public VParallelPanel(Frame frame, DataManager dataManager) {
		super(frame);

		Objects.requireNonNull(dataManager);

		this.dataManager = dataManager;
		this.dataTable = dataManager.getDataTable();

		dataManager.addUpdateListener(this);

		ToolTipManager toolTipManager = ToolTipManager.sharedInstance();
		toolTipManager.setInitialDelay(0);
		toolTipManager.setDismissDelay(Integer.MAX_VALUE);
		toolTipManager.setLightWeightPopupEnabled(false);
		toolTipManager.registerComponent(this);

		addMouseMotionListener(new MyMouseMotionListener());
		addMouseListener(new MyMouseListener());

		/* Initialize axises. */
		int attributeCount = dataTable.getAttributeCount();
		List<Integer> numberAttributeList = new LinkedList<>();

		for (int i = 0; i < attributeCount; i++) {
			if (Number.class.isAssignableFrom(dataTable.getAttributeType(i))) {
				numberAttributeList.add(i);
			}
		}

		int numberAttributeCount = numberAttributeList.size();
		int[] numberAttributeArray = new int[numberAttributeCount];
		axisMappers = new CoordinateMapper[numberAttributeCount];

		for (int i = 0; i < numberAttributeCount; i++) {
			numberAttributeArray[i] = numberAttributeList.get(i);
			axisMappers[i] = new CoordinateMapper();
		}

		config.init(dataTable, numberAttributeArray);
		colorConfig = config.getColor();
	}


	@Override
	public void updatePerformed() {
		repaint();
	}


	@Override
	public void destroyPerformed() {
		dataManager.removeUpdateListener(this);
	}


	/*
	 * Draws the graph.
	 */
	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);

		Dimension panelSize = getSize();
		Rectangle plotSize = new Rectangle(0, 5, panelSize.width, panelSize.height - 5);

		int axisCount = config.getAxisCount();

		g.setColor(ColorDefiner.BACKGROUND);
		g.fillRect(0, 0, panelSize.width, panelSize.height);
		g.setColor(ColorDefiner.NEUTRAL);

		xMapper.init(plotSize.x + 5, plotSize.width - 5, 0, axisCount - 1);
		colorMapper.init(0, 1, colorConfig.getCurrentMin(), colorConfig.getCurrentMax());

		rulerUtil.init(g, plotSize, dataTable, config, xMapper);
		rulerUtil.drawRulers();

		for (int i = 0; i < axisCount; i++) {
			AttributeConfig axisConfig = config.getAxis(i);
			axisMappers[i].init(plotSize.y, plotSize.getHeight(), axisConfig.getCurrentMax(), axisConfig.getCurrentMin());
		}

		rebuildLineList();
		drawLines(g);
	}


	/**
	 * Draws the markers.
	 */
	private void drawLines(Graphics g) {
		for (Map.Entry<Line, Set<DataElement>> entry : lines.entrySet()) {
			Line line = entry.getKey();

			Color color = ColorDefiner.get(
					dataManager,
					colorMapper,
					colorConfig,
					entry.getValue()
			);

			g.setColor(color);
			g.drawLine(line.getX1(), line.getY1(), line.getX2(), line.getY2());
		}
	}


	/**
	 * Rebuilds the map of markers within the plot.
	 */
	private void rebuildLineList() {
		int elementCount = dataTable.getElementCount();
		int axisTop = config.getAxisCount() - 1;

		lines.clear();

		for (int i = 0; i < elementCount; i++) {
			DataElement element = dataTable.getElement(i);

			if (!dataManager.isVisible(element)) {
				continue;
			}

			for (int j = 0; j < axisTop; j++) {
				double leftValue = element.getDouble(config.getAxis(j).getAttribute());
				double rightValue = element.getDouble(config.getAxis(j + 1).getAttribute());

				Line line = new Line(
						(int) Math.round(xMapper.toPixel(j)),
						(int) Math.round(axisMappers[j].toPixel(leftValue)),
						(int) Math.round(xMapper.toPixel(j + 1)),
						(int) Math.round(axisMappers[j + 1].toPixel(rightValue))
				);

				Set<DataElement> elementList = lines.get(line);

				if (elementList == null) {
					elementList = new HashSet<>();
					lines.put(line, elementList);
				}

				elementList.add(element);
			}
		}
	}


	/**
	 * @return The line touched by the point.
	 */
	private Line getTouchedLine(Point p) {
		double minDistance = Double.MAX_VALUE;
		Line closest = null;

		for (Line line : lines.keySet()) {
			Point linePoint1 = new Point(line.getX1(), line.getY1());
			Point linePoint2 = new Point(line.getX2(), line.getY2());
			double distance = linePoint1.distance(p) + linePoint2.distance(p) - linePoint1.distance(linePoint2);

			if (distance < minDistance) {
				minDistance = distance;
				closest = line;
			}
		}

		if (minDistance < 1) {
			return closest;
		} else {
			return null;
		}
	}


	/**
	 * Used for touching and panning.
	 */
	private class MyMouseMotionListener extends MouseMotionAdapter {
		@Override
		public void mouseMoved(MouseEvent e) {
			Line line = getTouchedLine(e.getPoint());
			Set<DataElement> elements;

			if (line != null) {
				elements = lines.get(line);
			} else {
				elements = new HashSet<>(0);
			}

			dataManager.setTouched(elements);
			setToolTipText(DataElementUtil.getElementTable(dataTable, elements));
		}
	}


	/**
	 * Used for highlighting, zooming and bringing up the menu.
	 */
	private class MyMouseListener extends MouseAdapter {
		@Override
		public void mouseClicked(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON1) {
				Line line = getTouchedLine(e.getPoint());
				Set<DataElement> elements;

				if (line != null) {
					elements = lines.get(line);
				} else {
					elements = new HashSet<>(0);
				}

				if (e.isControlDown()) {
					dataManager.swapHighlight(elements);
				} else {
					dataManager.setHighlights(elements);
				}

			} else if (e.getButton() == MouseEvent.BUTTON3) {
				JDialog dialog = new VParallelConfigDialog(frame, dataManager, config);
				dialog.addWindowListener(new DialogClosedListener());
				dialog.setVisible(true);
			}
		}
	}


	/**
	 * Called when dialogs are closed.
	 */
	private class DialogClosedListener extends WindowAdapter {
		@Override
		public void windowClosed(WindowEvent e) {
			axisMappers = new CoordinateMapper[config.getAxisCount()];

			for (int i = 0; i < axisMappers.length; i++) {
				axisMappers[i] = new CoordinateMapper();
			}

			repaint();
		}
	}
}
