/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import javax.swing.JDialog;
import javax.swing.ToolTipManager;

import vispaca.data.DataElement;
import vispaca.data.DataManager;
import vispaca.data.DataTable;
import vispaca.event.UpdateListener;
import vispaca.misc.AttributeConfig;
import vispaca.misc.ColorDefiner;
import vispaca.misc.CoordinateMapper;
import vispaca.misc.DataElementUtil;
import vispaca.misc.ScatterConfig;
import vispaca.misc.ScatterRulerUtil;


/**
 * Panel displaying a scatter plot.
 */
public class VScatterPanel extends VPanel implements UpdateListener {

	private static final long serialVersionUID = 7753018528178808238L;

	private Map<Rectangle, List<DataElement>> markers = new HashMap<>();

	private CoordinateMapper xMapper = new CoordinateMapper();
	private CoordinateMapper yMapper = new CoordinateMapper();
	private CoordinateMapper colorMapper = new CoordinateMapper();

	private ScatterRulerUtil rulerUtil = new ScatterRulerUtil();

	private ScatterConfig config = new ScatterConfig();

	private AttributeConfig xConfig;
	private AttributeConfig yConfig;
	private AttributeConfig colorConfig;

	private DataManager dataManager;
	private DataTable dataTable;


	public VScatterPanel(Frame frame, DataManager dataManager) {
		super(frame);

		Objects.requireNonNull(dataManager);

		this.dataManager = dataManager;
		this.dataTable = dataManager.getDataTable();

		dataManager.addUpdateListener(this);

		xConfig = config.getX();
		yConfig = config.getY();
		colorConfig = config.getColor();

		ToolTipManager toolTipManager = ToolTipManager.sharedInstance();
		toolTipManager.setInitialDelay(0);
		toolTipManager.setDismissDelay(Integer.MAX_VALUE);
		toolTipManager.setLightWeightPopupEnabled(false);
		toolTipManager.registerComponent(this);

		addMouseWheelListener(new MyMouseWheelListener());
		addMouseMotionListener(new MyMouseMotionListener());
		addMouseListener(new MyMouseListener());
	}


	@Override
	public void updatePerformed() {
		repaint();
	}


	@Override
	public void destroyPerformed() {
		dataManager.removeUpdateListener(this);
	}


	/*
	 * Draws the graph.
	 */
	@Override
	protected void paintComponent(Graphics g) {
		super.paintComponent(g);

		Dimension size = getSize();

		g.setColor(ColorDefiner.BACKGROUND);
		g.fillRect(0, 0, size.width, size.height);
		g.setColor(ColorDefiner.NEUTRAL);

		xMapper.init(0, size.width - 6, xConfig.getCurrentMin(), xConfig.getCurrentMax());
		yMapper.init(5, size.height, yConfig.getCurrentMax(), yConfig.getCurrentMin());
		colorMapper.init(0, 1, colorConfig.getCurrentMin(), colorConfig.getCurrentMax());

		rulerUtil.init(g, xMapper, yMapper);
		rulerUtil.drawRulers(9, 9);

		rebuildMarkerList();
		drawMarkers(g);
	}


	/**
	 * Draws the markers.
	 */
	private void drawMarkers(Graphics g) {
		for (Map.Entry<Rectangle, List<DataElement>> entry : markers.entrySet()) {
			Rectangle marker = entry.getKey();

			Color color = ColorDefiner.get(
					dataManager,
					colorMapper,
					colorConfig,
					entry.getValue()
			);

			g.setColor(color);
			g.fillRect(marker.x + 1, marker.y + 1, marker.width - 2, marker.height - 2);
			g.setColor(ColorDefiner.NEUTRAL);
			g.drawRect(marker.x, marker.y, marker.width - 1, marker.height - 1);
		}
	}


	/**
	 * Rebuilds the map of markers within the plot.
	 */
	private void rebuildMarkerList() {
		int elementCount = dataTable.getElementCount();

		boolean xVisible = 0 <= xConfig.getAttribute();
		boolean yVisible = 0 <= yConfig.getAttribute();

		if (!xVisible && !yVisible) {
			return;
		}

		double xValueLow = Math.min(xMapper.getValueMin(), xMapper.getValueMax());
		double xValueHigh = Math.max(xMapper.getValueMin(), xMapper.getValueMax());
		double yValueLow = Math.min(yMapper.getValueMin(), yMapper.getValueMax());
		double yValueHigh = Math.max(yMapper.getValueMin(), yMapper.getValueMax());

		markers.clear();

		for (int i = 0; i < elementCount; i++) {
			DataElement element = dataTable.getElement(i);

			if (!dataManager.isVisible(element)) {
				continue;
			}

			double xValue = 0;
			double yValue = 0;

			if (xVisible) {
				xValue = element.getDouble(xConfig.getAttribute());

				if (xValue < xValueLow || xValueHigh < xValue) {
					continue;
				}
			}

			if (yVisible) {
				yValue = element.getDouble(yConfig.getAttribute());

				if (yValue < yValueLow || yValueHigh < yValue) {
					continue;
				}
			}

			int yPixel = (int) Math.round(yMapper.toPixel(yValue));
			int xPixel = (int) Math.round(xMapper.toPixel(xValue));

			Rectangle marker = new Rectangle(xPixel - 3, yPixel - 3, 7, 7);

			List<DataElement> markerList = markers.get(marker);

			if (markerList == null) {
				markerList = new LinkedList<>();
				markers.put(marker, markerList);
			}

			markerList.add(element);
		}
	}


	/**
	 * Used for touching and panning.
	 */
	private class MyMouseMotionListener extends MouseMotionAdapter {
		private Point previous;

		@Override
		public void mouseMoved(MouseEvent e) {
			Point point = e.getPoint();
			Set<DataElement> elements = new HashSet<>();

			previous = point;

			for (Rectangle rectangle : markers.keySet()) {
				if (rectangle.contains(point)) {
					elements.addAll(markers.get(rectangle));
				}
			}

			dataManager.setTouched(elements);

			if (elements.isEmpty()) {
				setToolTipText(null);
			} else {
				setToolTipText(DataElementUtil.getElementTable(dataTable, elements));
			}
		}

		@Override
		public void mouseDragged(MouseEvent e) {
			Point point = e.getPoint();

			if ((e.getModifiersEx() & MouseEvent.BUTTON2_DOWN_MASK) != 0 && previous != null) {
				double prevXValue = xMapper.toValue(previous.getX());
				double prevYValue = yMapper.toValue(previous.getY());
				double curXValue = xMapper.toValue(point.getX());
				double curYValue = yMapper.toValue(point.getY());

				double xAdjust = curXValue - prevXValue;
				double yAdjust = curYValue - prevYValue;

				if (xConfig.getCurrentMin() < xConfig.getCurrentMax()) {
					xConfig.setCurrentMin(xConfig.getCurrentMin() - xAdjust);
					xConfig.setCurrentMax(xConfig.getCurrentMax() - xAdjust);
				} else {
					xConfig.setCurrentMin(xConfig.getCurrentMin() + xAdjust);
					xConfig.setCurrentMax(xConfig.getCurrentMax() + xAdjust);
				}

				if (yConfig.getCurrentMin() < yConfig.getCurrentMax()) {
					yConfig.setCurrentMin(yConfig.getCurrentMin() - yAdjust);
					yConfig.setCurrentMax(yConfig.getCurrentMax() - yAdjust);
				} else {
					yConfig.setCurrentMin(yConfig.getCurrentMin() + yAdjust);
					yConfig.setCurrentMax(yConfig.getCurrentMax() + yAdjust);
				}

				repaint();
			}

			previous = point;
		}
	}


	/**
	 * Used for highlighting, zooming and bringing up the menu.
	 */
	private class MyMouseListener extends MouseAdapter {
		@Override
		public void mouseClicked(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON1) {
				Point point = e.getPoint();
				Set<DataElement> elements = new HashSet<>();

				for (Rectangle rectangle : markers.keySet()) {
					if (rectangle.contains(point)) {
						elements.addAll(markers.get(rectangle));
					}
				}

				if (e.isControlDown()) {
					dataManager.swapHighlight(elements);
				} else {
					dataManager.setHighlights(elements);
				}

			} else if (e.getButton() == MouseEvent.BUTTON3) {
				JDialog dialog = new VScatterConfigDialog(frame, dataManager, config);
				dialog.addWindowListener(new DialogClosedListener());
				dialog.setVisible(true);
			}
		}
	}


	/**
	 * Used for zooming.
	 */
	private class MyMouseWheelListener implements MouseWheelListener {
		@Override
		public void mouseWheelMoved(MouseWheelEvent e) {
			if (e.getWheelRotation() == 0) {
				return;
			}

			boolean positive = e.getWheelRotation() < 0;
			boolean xEnable = xMapper.getPixelMin() <= e.getX() && e.getX() <= xMapper.getPixelMax() && 0 < xConfig.getAttribute();
			boolean yEnable = yMapper.getPixelMin() <= e.getY() && e.getY() <= yMapper.getPixelMax() && 0 < yConfig.getAttribute();

			if (xEnable) {
				performZoon(positive, xConfig);
			}

			if (yEnable) {
				performZoon(positive, yConfig);
			}

			repaint();
		}

		private void performZoon(boolean positive, AttributeConfig config) {
			double effect;

			if (config.getCurrentMin() < config.getCurrentMax()) {
				effect = (config.getCurrentMax() - config.getCurrentMin()) * 0.1;
				if (positive) {
					config.setCurrentMin(config.getCurrentMin() + effect);
					config.setCurrentMax(config.getCurrentMax() - effect);
				} else {
					config.setCurrentMin(config.getCurrentMin() - effect);
					config.setCurrentMax(config.getCurrentMax() + effect);
				}

			} else {
				effect = (config.getCurrentMin() - config.getCurrentMax()) * 0.1;
				if (positive) {
					config.setCurrentMin(config.getCurrentMin() - effect);
					config.setCurrentMax(config.getCurrentMax() + effect);
				} else {
					config.setCurrentMin(config.getCurrentMin() + effect);
					config.setCurrentMax(config.getCurrentMax() - effect);
				}
			}
		}
	}


	/**
	 * Called when dialogs are closed.
	 */
	private class DialogClosedListener extends WindowAdapter {
		@Override
		public void windowClosed(WindowEvent e) {
			repaint();
		}
	}
}
