/*
 * Copyright (C) 2013  Joakim Soderlund
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


package vispaca.misc;


/**
 * Maps values to pixel and vice versa.
 *
 * Note that this class will not function properly if pixelMin is larger than
 * pixelMax. However, values where valueMin is larger than valueMax are
 * supported and can be used to invert an axis.
 */
public class CoordinateMapper {

	private double pixelMin;
	private double pixelMax;
	private double valueMin;
	private double valueMax;

	private double pixelDelta;
	private double valueDelta;


	public CoordinateMapper() {

	}


	public CoordinateMapper(double pixelMin, double pixelMax, double valueMin, double valueMax) {
		init(pixelMin, pixelMax, valueMin, valueMax);
	}


	/**
	 * For late initialization or reuse of objects.
	 */
	public void init(double pixelMin, double pixelMax, double valueMin, double valueMax) {
		if (pixelMax <= pixelMin) {
			throw new IllegalArgumentException("Pixel area must be larger than 0.");
		}

		this.pixelMin = pixelMin;
		this.pixelMax = pixelMax;
		this.valueMin = valueMin;
		this.valueMax = valueMax;

		pixelDelta = pixelMax - pixelMin;
		valueDelta = Math.abs(valueMax - valueMin);
	}


	/**
	 * Maps a value to a pixel.
	 *
	 * @param value The value to map.
	 * @return The corresponding pixel.
	 */
	public double toPixel(double value) {
		if (value < Math.min(valueMin, valueMax)) {
			return pixelMin;
		} else if (Math.max(valueMin, valueMax) < value) {
			return pixelMax;
		} else {
			double ratio = Math.abs(value - valueMin) / valueDelta;
			return (ratio * pixelDelta) + pixelMin;
		}
	}


	/**
	 * Maps a pixel to a value.
	 *
	 * @param pixel The pixel to map.
	 * @return The corresponding value.
	 */
	public double toValue(double pixel) {
		if (pixel < Math.min(pixelMin, pixelMax)) {
			return valueMin;
		} else if (Math.max(pixelMin, pixelMax) < pixel) {
			return valueMax;
		} else {
			double ratio = Math.abs(pixel - pixelMin) / pixelDelta;

			if (valueMin < valueMax) {
				return (ratio * valueDelta) + valueMin;
			} else {
				return valueMin - (ratio * valueDelta);
			}
		}
	}


	/**
	 * @return The length of the coordinate system.
	 */
	public double getPixelDelta() {
		return pixelDelta;
	}


	/**
	 * @return The value range of the coordinate system.
	 */
	public double getValueDelta() {
		return valueDelta;
	}


	public double getPixelMin() {
		return pixelMin;
	}


	public void setPixelMin(double pixelMin) {
		this.pixelMin = pixelMin;
		pixelDelta = Math.abs(pixelMax - pixelMin);
	}


	public double getPixelMax() {
		return pixelMax;
	}


	public void setPixelMax(double pixelMax) {
		this.pixelMax = pixelMax;
		pixelDelta = Math.abs(pixelMax - pixelMin);
	}


	public double getValueMin() {
		return valueMin;
	}


	public void setValueMin(double valueMin) {
		this.valueMin = valueMin;
		valueDelta = Math.abs(valueMax - valueMin);
	}


	public double getValueMax() {
		return valueMax;
	}


	public void setValueMax(double valueMax) {
		this.valueMax = valueMax;
		valueDelta = Math.abs(valueMax - valueMin);
	}
}
